-- Hypatia
-- version 4.5

-- global procedure write - writes to screen and/or log file
-- global function magntd - returns magnitude
-- global function rround - replaces round()
-- global function roundsig - rounds x to n significant digits
-- global function snapinteger - rounds values close to integer (but not close to 0) to integer
-- global function round0 - rounds absolute values below zero threshold to 0
-- global function isnat - returns true if number is zero or positive integer
-- procedure getfmt - get numeric output format parameter
-- global function sprint15 -- replacement for sprint
-- global function fmtnumber - format number according to current format settings
-- global function fmtresult - calls fmtnumber
-- global procedure dispmode - displays program mode settings
-- global procedure dispvariables - displays some or all variables
-- global procedure dispudes - displays some or all UDEs
-- function badfilename - verifies file name of SAVE commands
-- global procedure savevariables - saves all variables to a file
-- global procedure savebuffer - saves content of buffer to a file
-- global procedure saveudes - saves all user defined objects to a file
-- global procedure storeval - assigns value of $ to variable with STO command, stores it in variable list
-- global procedure storeval2 - as above, but for $var = value or expression command
-- global procedure storeude - assigns value to user defined element, stores it in ude list
-- global procedure delvar - deletes variable
-- global procedure delude - deletes user defined element
-- global function ccopy - copies content of file to clipboard
-- global function notmute - returns true if not loop, or if loop with ? option
-- global function ndrand - normal distributed random number
-- global procedure clearhy - clears file hy, or adds line break
-- global procedure savehy - saves result to file hy
-- global procedure savehyin - saves input line to file hyin
-- global procedure createini - creates empty hy.ini file if it doesn't exist
-- global procedure expand - inserts () files and user defined elements into input line


include graphics.e

global procedure write(integer out, object w)
	-- writes string w to screen and/or log file
	-- out = 1: only screen, out = 2 screen and log file, out = 0 only log file
	-- write(1, text) same as disp(text)
	-- do not write to log file when mode_log = 0
	-- \n is always added at the end of the output string

	integer D -- hy.log file number

	if out > 0 then
		if flag_clc = 0 then disp(w) end if
	end if

	if out = 0 or out = 2 then
		if mode_log then
			D = open(hdir & "hy.log", "a")
			if D = -1 then
				disp("Warning: log file hy.log could not be opened for writing, LOG mode set to OFF")
				mode_log = 0
			else
				puts(D, w & "\n")
				close(D)
			end if
		end if
	end if
end procedure


global function magntd(atom x)
	-- returns exponent part of of x in scientific notation
	integer d
	if x = 0 then return 0 end if

	x = abs(x)
	d = floor(log10(x))                 -- exponent
	if x >= apower(10, d+1) then        -- correct d in case log10 is inaccurate for values very close to powers of 10
		d += 1
	elsif x < apower(10, d) then
		d -= 1
	end if
	return d
end function


global function rround(atom x, integer d)
	atom y, c
	integer m, mdc

-- m    magnitude plus 1 (= digits before dec.p.)
-- d    digits to round, pos. or 0 is after dec.p., neg. is set 0 before dec.p.
--      when total length of rounded result > 17 digits, then d gets reduced, but not below 0
-- mdc  to calculate c
-- c    correction element added to 0.5: 0. ...1, so that 1 is added at the 18th digit of orig. argument, but not larger than 0.001
-- rounding to n digits after dec. point:

	if mode_round then                         -- rounding mode is "Phix"
		x = round(x, apower(10, d))
		return x
	end if

	m = magntd(x) + 1                          -- number of dig. before dec.p. (0 for 0.1 etc.)
	if m + d < 0 then return 0 end if          -- magnitude too low for rounding to n digits (applies to pos. and neg. d)

	mdc = m - 18                               -- 1 is added to 18th digit of argument (not 18th digit after dec.p.)
	if mdc < -2 then                              -- but max. 0.001
		c = apower(10, mdc)
	else
		c = 0.001
	end if

	if d >= 0 then                             -- rounding to n digits after dec. point
		if integer(x) then return x end if          -- integers cannot have decimal digits rounded

		if m + d > 17 then                          -- dig. before dec.p. + rounded dig. after dec.p.: max. allowed 17
			d = 17 - m
			if d < 0 then d = 0 end if                -- dig. after dec.p. not less than 0 (no rounding before dec.p.)
		end if
		y = abs(x) + c
		y = y * apower(10, d) + 0.5
		y = trunc(y) / apower(10, d) * sign(x)
	else                                       -- rounding to n 0s before dec. point
		d = -d
		y = abs(x) + c
		y = y / apower(10, d) + 0.5
		y = trunc(y) * apower(10, d) * sign(x)
	end if
	return y
end function


global function roundsig(atom x, integer n)
  -- rounds x to n significant digits
  integer d

	if x = 0 then return 0 end if

  d = magntd(x) + 1 - n        -- digits to move dec. point to the left
  if d < 0 then
	  x = x * power(10, -d)      -- only use positive powers of 10!
		x = rround(x, 0)
		x = x / power(10, -d)
	else
		x = x / power(10, d)
		x = rround(x, 0)
		x = x * power(10, d)
	end if
	return x
end function


global function snapinteger(atom x)
	-- rounds numbers close to integer (but not close to 0) to integer
	-- only for values < 1E15 (max. 15 digits)
	-- called by done() only if mode_integerbias
	atom y
	if integer(x) then return x end if               -- is integer
	if abs(x) < 0.999999999 then return x end if     -- certainly too small to be rounded to 1
	if abs(x) >= 1E15 then return x end if           -- more than 15 digits before dec.p.

	y = roundsig(x, 17)                              -- round to 17 significant digits
	if integer(y) then
		return y
	else
		return x
	end if
end function


global function round0(atom w)
	-- rounds absolute values below zero threshold to zero

	if zerogate < 0 then
		err = 1
		errtxt = "zero threshold variable $zero is negative"
		return 0
	elsif zerogate = 0 then
		return w
	else
		if abs(w) < zerogate then
			return 0
		else
			return w
		end if
	end if
end function


global function isnat(atom x)
	-- returns 1 if x is zero or positive integer
	if x < 0 then
		return 0
	else
		return integer(x)
	end if
end function


global procedure getfmt()
	-- get numeric output format
	object v
	if length(stack) = 1 then
		fmt = -99
	elsif length(stack) > 2 then
		err = 1
		errtxt = stack[1] & " can only have one parameter"
	else
		v = readnumber(stack[2], 1)
		if atom(v) then
			if equal(stack[1], "fhex") then
				if v < 2 or v > 12 then
					err = 1
					errtxt = "number of hexadecimal digits must be 2 to 12"
				else
					fmt = v
				end if
			elsif equal(stack[1], "fbin") then
				if v < 2 or v > 48 then
					err = 1
					errtxt = "number of binary digits must be 2 to 48"
				else
					fmt = v
				end if
			else                             -- fdec
				if v < -15 or v > 15 then
					err = 1
					if v > 0 then
						errtxt = "number of digits after decimal point must be 0 to 15"
					else
						errtxt = "min. number of digits before decimal point must be -15 to -1"
					end if
				else
					fmt = v
				end if
			end if
		else
			err = 1
			errtxt = stack[2] & " is not a valid number of digits"
		end if
	end if
end procedure


-- a substitute for sprint(), returns up to 18 digits instead of 10
-- like sprint(), switches to scientific (exponential) notation for small and large numbers,
-- but unlike sprint() still returns up to 15 significant digits after (up to 3) zeros after decimal point for absolute values < 1
-- digs = number of digits, default = 15
-- 15 digits get changed to 18 for integer numbers
global function sprint15(atom a, integer digs = 15)
	string f, g
	integer flag_neg = 0
	integer pos_e
	integer xp, dp, ixs
	string fx

	if a = 0 then
		return "0"
	elsif a > 0 then
		flag_neg = 0
	else
		a = -a
		flag_neg = 1
	end if
	
	if digs < 3 or digs > 18 then return "digits out of range" end if
	
	if digs = 15 and integer(a) then digs = 18 end if

	fx = "%1." & sprint(digs - 1) & "e"

	f = sprintf(fx, a)               -- creates string "1.23456789012345exx"

	pos_e = match('e', f)               -- position of 'e' in string f (should be digs + 2 - there was a bug in Phix up to 1.0.2, fixed in 1.0.3)
	
	if pos_e != digs + 2 then                 -- to be on the safe side (see above)
		return sprint(a)
	end if
	
	g = f[1] & f[3..digs + 1]                 -- creates string "123456789012345"
	xp = readnumber(f[pos_e+1..$])            -- value of exponent

	if xp < -4 or xp >= digs then             -- switch to scientific
		g = f[1..digs + 1]
		while g[$] = '0' do
			g = g[1..$-1]
		end while
		if g[$] = '.' then g = g[1..$-1] end if
		g = g & f[pos_e..$]
	else
		if xp < 0 then
			g = "0." & repeat('0', -xp-1) & g
			while g[$] = '0' do
				g = g[1..$-1]
			end while
		else
			g = g[1..xp+1] & '.' & g[xp+2..$]
			while g[$] = '0' do
				g = g[1..$-1]
			end while
			if g[$] = '.' then g = g[1..$-1] end if
		end if
	
		if fmt_apostr then
			dp = find('.', g)
			if dp = 0 then dp = length(g) + 1 end if
			ixs = dp - 3
			while ixs > 1 do
				g = g[1..ixs-1] & "'" & g[ixs..$]
				ixs -= 3
			end while
		end if
	end if

	if flag_neg then g = '-' & g end if
	return g
end function


global function fmtnumber(atom r)
  -- formats number according to current format settings

	atom b, bb
	object rf
	object rr
	integer ixs

	if mode_dhb = 0 then               -- decimal
		if fmt = -99 then
			rr = sprint15(r, fmt_digits)
		else
			integer dp
			integer fmtx
			if fmt < 0 then                   -- with leading zeros
				integer fmt_ap                     -- sprint15 has to be called with apostrophes OFF
				fmt_ap = fmt_apostr
				fmt_apostr = 0
				fmtx = -fmt
				rr = sprint15(abs(r), fmt_digits)
				fmt_apostr = fmt_ap
				dp = match('.', rr)
				if dp then
					if dp <= fmtx then rr = repeat('0', fmtx - dp + 1) & rr end if
				else
					if length(rr) < fmtx then rr = repeat('0', fmtx - length(rr)) & rr end if
				end if
				if fmt_apostr then
					dp = find('.', rr)
					if dp = 0 then dp = length(rr) + 1 end if
					ixs = dp - 3
					while ixs > 1 do
						rr = rr[1..ixs-1] & "'" & rr[ixs..$]
						ixs -= 3
					end while
				end if
				if sign(r) >= 0 then
					rr = " " & rr
				else
					rr = "-" & rr
				end if
			else                              -- with fixed number of decimal digits (can be 0)
				dp = magntd(r)
				if dp > 14 then
					rr = sprintf("%e", r)
				else
					if fmt + dp > 14 then                 -- show only 15 digits
						fmtx = 14 - dp
					else
						fmtx = fmt
					end if
					r = rround(r, fmtx)                   -- sprint %f doesn't round reliably
					rf = "%1." & sprint(fmtx) & "f"
					rr = trim(sprintf(rf, r))
				end if
				if fmt_apostr then
					dp = find('.', rr)
					if dp = 0 then dp = length(rr) + 1 end if
					ixs = dp - 3
					while ixs > 1 do
						rr = rr[1..ixs-1] & "'" & rr[ixs..$]
						ixs -= 3
					end while
				end if
			end if
		end if
		return rr
	elsif mode_dhb = 1 then                       -- hexadecimal
		if isnat(r) and r < 281474976710656 then
			rf = sprintf("%x", r)
			if length(rf) < fmt then
				rf = repeat('0', fmt - length(rf)) & rf
			end if
			if fmt_apostr then
				ixs = length(rf) - 4
				while ixs > 0 do
					rf = rf[1..ixs] & "'" & rf[ixs+1..$]
					ixs -= 4
				end while
			end if
			return "&h" & rf
		else
			return sprint15(r)
		end if
	elsif mode_dhb = 2 then                      -- binary
		if isnat(r) and r < 281474976710656 then
			rf = ""
			b = 0
			bb = 1
			while bb * 2 <= r do
				b = b + 1
				bb = bb * 2
			end while
			for i = b to 0 by -1 do
				if r < bb then
					rf = rf & "0"
				else
					rf = rf & "1"
					r = r - bb
				end if
				bb = bb / 2
			end for
			if length(rf) < fmt then
				rf = repeat('0', fmt - length(rf)) & rf
			end if
			if fmt_apostr then
				ixs = length(rf) - 8
				while ixs > 0 do
					rf = rf[1..ixs] & "'" & rf[ixs+1..$]
					ixs -= 8
				end while
			end if
			return "&b" & rf
		else
			return sprint15(r, fmt_digits)
		end if
	elsif mode_dhb = 3 then                   -- scientific
		if fmt_digits = 9 then
			return sprintf("%1.8e", r)
		else
			return sprintf("%1.14e", r)
		end if
	elsif mode_dhb = 4 then                   -- lakh/crore
		if abs(r) < 1E5 then
			return sprint15(r)
		elsif abs(r) < 1E7 then
			return sprint15(r/1E5, fmt_digits) & " lakh"
		else
			return sprint15(r/1E7, fmt_digits) & " crore"
		end if
	elsif mode_dhb = 5 then                   -- million
		if abs(r) < 1E6 then
			return sprint15(r, fmt_digits)
		else
			return sprint15(r/1E6, fmt_digits) & " million"
		end if
	end if
end function


global function fmtresult()
  -- formats result according to current format settings
	atom r
	r = var[1][2]                      -- calculation result
	return fmtnumber(r)
end function


global procedure dispmode()
	integer mode_dhb_orig, fmt_apostr_orig
	disp("For help use HELP command\n")
	if length(hdir) = 0 then
		disp("I/O file location: current folder " & get_proper_dir("."))
	else
		disp("I/O file location: program folder " & hdir)
	end if
	disp("External editor: " & exteditor)

	if flag_x then
		disp("Current screen width not available under Linux")                   -- *** video_config currently doesn't work with Linux ***
	else
		disp("Current screen width: " & sprint(video_config()[10]) & " characters")
	end if

	if mode_dhb = 0 then
		puts(1, "Result format: decimal")
		if fmt != -99 then
			if fmt >=0 then
				puts(1, ", " & sprint(fmt) & " digits after decimal point")
			else
				puts(1, ", " & sprint(-fmt) & " zeros before decimal point")
			end if
		end if
		disp("")
	elsif mode_dhb = 1 then
		puts(1, "Result format: hexadecimal")
		if fmt > 1 then
			disp(", min. " & sprint(fmt) & " digits")
		else
			disp("")
		end if
	elsif mode_dhb = 2 then
		puts(1, "Result format: binary")
		if fmt > 1 then
			disp(", min. " & sprint(fmt) & " digits")
		else
			disp("")
		end if
	elsif mode_dhb = 3 then
		disp("Result format: scientific")
	elsif mode_dhb = 4 then
		disp("Result format: lakh/crore")
	elsif mode_dhb = 5 then
		disp("Result format: million")
	end if
	
	if fmt_digits = 9 then
		disp("Digits shown: 9 (short)")
	else
		disp("Digits shown: 15 (long)")
	end if

	if mode_autoinc then
		disp("Auto include $: ON")
	else
		disp("Auto include $: OFF")
	end if
	if mode_deg then
		disp("Angle mode: degrees")
	else
		disp("Angle mode: radians")
	end if

	disp("Zero threshold: " & sprint15(zerogate))

	if mode_integerbias then
		disp("Integer bias: ON")
	else
		disp("Integer bias: OFF")
	end if
	
	if mode_round then
		disp("Hypatia rounding mode: disabled")
	end if

	mode_dhb_orig = mode_dhb                 -- display max. number of passes in dec. format with apostrophes, independent of format settings
	fmt_apostr_orig = fmt_apostr
	mode_dhb = 0
	fmt_apostr = 1
	disp("Max. number of loop passes: " & fmtnumber(max_repeat))
	mode_dhb = mode_dhb_orig
	fmt_apostr = fmt_apostr_orig
	
	if mode_echo then
		disp("Echo mode: ON")
	else
		disp("Echo mode: OFF")
	end if
	if mode_copy then
		disp("Copy all results to clipboard: ON")
	else
		disp("Copy all results to clipboard: OFF")
	end if
	if mode_log then
		disp("Logging: ON")
	else
		disp("Logging: OFF")
	end if
	if flag_buffer then
		if length(buffer) then
			disp("Buffer mode: ON")
		else
			disp("Buffer mode: ON, buffer is empty")
		end if
	else
		disp("Buffer mode: OFF")
	end if

	disp("")
end procedure


global procedure dispvariables()
	-- displays list of variables, and "Angle mode is degrees" if that mode has been set (radians is default)
	integer found

	if length(stack) = 1 then
		if mode_deg then write(2, "Angle mode is degrees") end if
		for j = 3 to length(var) do
			write(2, var[j][1] & " = " & sprint15(var[j][2], fmt_digits))
		end for
		write(2, "$$ = " & sprint15(var[2][2], fmt_digits))
		write(2, "$ = " & sprint15(var[1][2], fmt_digits))
		write(2, "")
	else
		for s = 2 to length(stack) do
			if stack[s][1] = '$' then
				found = 0
				for j = 1 to length(var) do
					if equal(stack[s], var[j][1]) then
						write(2, var[j][1] & " = " & sprint15(var[j][2], fmt_digits))
						found = 1
						exit
					end if
				end for
				if found = 0 then write(2, stack[s] & " variable not found") end if
			elsif equal(stack[s], "i") then
				if flag_do then               -- do loop: i = loop index
					write(2, "I = " & sprint(i_repeat))
				else                          -- repeat loop: I = loop index + 1
					write(2, "I = " & sprint(i_repeat + 1))
				end if
			elsif equal(stack[s], "time") then
				if sign(num_repeat) then
					write(2, "TIME = " & sprint(time() - time_start))
				else
					write(2, "TIME = 0, not a loop")
				end if
			else
				write(2, stack[s] & " is not a valid variable name")
			end if
		end for
	end if
end procedure


global procedure dispudes()
	-- displays list of user defined elements
	object wl -- line to be written
	integer cols -- display columns 
	integer found

	if flag_x then
		cols = 80                   -- *** video_config() currently doesn't work with Linux, 80 columns assumed ***
	else
		cols = video_config()[10]
	end if

	if length(stack) = 1 then
		if length(ude) = 0 then
			write(2, "There are no user defined elements")
		else
			for j = 1 to length(ude) do
				wl = atrim(ude[j][1]) & " = " & atrim(ude[j][2])
				if length(wl) > cols then wl = wl[1..cols-4] & "..." end if
				write(2, wl)
			end for
			write(2, "")
		end if
	else
		for s = 2 to length(stack) do
			if stack[s][1] = '@' then
				found = 0
				for j = 1 to length(ude) do
					if equal(stack[s], atrim(ude[j][1])) then
						wl = atrim(ude[j][1]) & " = " & atrim(ude[j][2])
						if length(wl) > cols then wl = wl[1..cols-4] & "..." end if
						write(2, wl)
						found = 1
						exit
					end if
				end for
				if found = 0 then write(2, stack[s] & " user defined element not found") end if
			else
				write(2, stack[s] & " is not a valid UDE name")
			end if
		end for
	end if
end procedure


function badfilename(object f)
	-- checks whether SAVE command file name is hy, hy.ini, hyin, hy.log, *.exe or *.pdf
	-- checks whether file name has ( ) * ?
	-- checks whether file is name of a folder
	-- asks whether to overwrite existing file
	object ff
	integer fex
	
	if equal(f, "hy") or equal(f, "hy.ini") or equal(f, "hyin") or equal(f, "hy.log") then
		err = 1
		errtxt = "file name " & f & " is not permitted"
	elsif find ('?', f) or find ('*', f) then
		err = 1
		errtxt = "invalid character ? or * in file name"
	elsif find('(', f) or find(')', f) then
		err = 1
		errtxt = "( ) not allowed in file name"
	elsif find ('\\', f) then
		err = 1
		errtxt = "\\ not allowed in file name"
	elsif find ('?', f) or find ('*', f) or find ('\"', f) or find ('/', f) or find ('|', f) or find ('<', f) or find ('>', f) then
		err = 1
		errtxt = "invalid character ? * / | < > or \" in file name"
	elsif f[1] = '@' or f[1] = '$' then
		err = 1
		errtxt = "file name must not begin with @ or $"
	else
		object fext
		fext = splitx(f, '.')[2]
		if equal(fext, "exe") or equal(fext, "dll") or equal(fext, "bin") or equal(fext, "pdf") then
			err = 1
			errtxt = "file extension ." & fext & " is not permitted"
		end if
	end if
	
	if err then return 1 end if

	ff = hdir & f
	fex = get_file_type(ff)
	if fex = 2 then
		err = 1
		errtxt = f & " is the name of a folder"
		return 1
	elsif fex = 1 then
		puts(1, "File " & f & " already exists. Overwrite? ")
		if yesno() = 0 then
			write(2, "Aborted, existing file not overwritten")
			return 1
		end if
	end if
	return 0
end function


global procedure savevariables(object f, object t)
	-- saves list of variables, and USE DEG if angle mode is degrees, to file
	-- does not save $ and $$
	integer D -- file number
	object ff -- full file name

	if badfilename(f) then return end if

	ff = hdir & f
	D = open(ff, "w")
	if D = -1 then
		err = 1
		errtxt = f & " could not be opened for writing"
	else
		if length(t) > 0 then puts(D, "# " & t & "\n") end if
		if mode_deg then puts(D, "USE DEG\n") end if
		for j = 3 to length(var) do
			puts(D, var[j][1] & " = " & sprint15(var[j][2], 18) & "\n")
		end for
		close (D)
		write(2, "Variables saved to file " & f)
		if mode_deg then write(2, "Line USE DEG included, restoring these variables will set angle mode to degrees") end if
	end if
end procedure


global procedure savebuffer(object f, object t)
	-- saves content of buffer to file
	integer D -- file number
	object ff -- full file name

	if badfilename(f) then return end if

	ff = hdir & f
	D = open(ff, "w")
	if D = -1 then
		err = 1
		write(2, "Error: file " & f & " could not be opened for writing")
	else
		if length(t) > 0 then puts(D, "# " & t & "\n") end if
		for j = 1 to length(buffer) do
			puts(D, buffer[j] & "\n")
		end for
		close (D)
		write(2, "Buffer saved to file " & f)
	end if
end procedure


global procedure saveudes(object f, object t)
	-- saves list of user defined elements to file
	integer D -- file number
	object ff -- full file name

	if badfilename(f) then return end if

	ff = hdir & f
	D = open(ff, "w")
	if D = -1 then
		err = 1
		write(2, "Error: file " & f & " could not be opened for writing")
	else
		if length(t) > 0 then puts(D, "# " & t & "\n") end if
		for j = 1 to length(ude) do
			puts(D, atrim(ude[j][1]) & " = " & atrim(ude[j][2]) & "\n")
		end for
		close (D)
		write(2, "User defined elements saved to file " & f)
	end if
end procedure


global procedure storeval()
	-- assigns value to variable, creates variable if it does not exist
	integer i

	if length(stack) != 2 then
		err = 1
		errtxt = "STO command needs one argument (variable name)"
	elsif stack[2][1] != '$' then
		err = 1
		errtxt = "invalid variable name, must start with $"
	elsif equal(stack[2], "$") or equal(stack[2], "$$") then
		err = 1
		errtxt = "variables $ and $$ may not be used in STO command"
	end if
	if err then return end if

	i = 0
	for j = 1 to length(var) do
		if equal(stack[2], var[j][1]) then
			var[j][2] = var[1][2]
			i = 1
			exit
		end if
	end for
	if not i then
		var = append(var, {"", ""})
		i = length(var)
		var[i][1] = stack[2]
		var[i][2]  = var[1][2]
	end if
	if equal(stack[2], "$zero") then zerogate = var[1][2] end if
end procedure


global procedure storeval2(object varname, atom varval)
	-- as storeval, but for alternative syntax $var = value or expression
	integer i
	
	i = 0
	for j = 1 to length(var) do
		if equal(varname, var[j][1]) then
			var[j][2] = varval
			i = 1
			exit
		end if
	end for
	if not i then
		var = append(var, {"", ""})
		i = length(var)
		var[i][1] = varname
		var[i][2] = varval
	end if
	if equal(varname, "$zero") then zerogate = varval end if
end procedure


global procedure storeude(object udename, object udeval)
	-- assigns value to user defined object, creates ude if it does not exist
	-- both names and values are stored with leading and trailing blanks
	integer i
	
	i = 0
	udename = " " & udename & " "
	udeval = " " & udeval & " "
	for j = 1 to length(ude) do
		if equal(udename, ude[j][1]) then
			ude[j][2] = udeval
			i = 1
			exit
		end if
	end for
	if not i then
		ude = append(ude, {"", ""})
		i = length(ude)
		ude[i][1] = udename
		ude[i][2] = udeval
	end if
end procedure


global procedure delvar()
	-- deletes variable
	integer i

	if length(stack) != 2 then
		err = 1
		errtxt = "DEL command needs one argument (variable name)"
	elsif stack[2][1] != '$' then
		err = 1
		errtxt = "invalid variable name (must start with $)"
	elsif equal(stack[2], "$") or equal(stack[2], "$$") then
		err = 1
		errtxt = "variables $ and $$ can not be deleted"
	end if
	if err then return end if

	i = 0
	for j = 1 to length(var) do
		if equal(stack[2], var[j][1]) then
			var = var[1..j-1] & var[j+1..$]
			i = 1
			exit
		end if
	end for
	if not i then
		err = 1
		errtxt = "variable " & stack[2] & " not found"
	end if
	if equal(stack[2], "$zero") then zerogate = def_zero end if
end procedure


global procedure delude()
	-- deletes user defined element
	integer i

	if length(stack) != 2 then
		err = 1
		errtxt = "DEL@ command needs one argument (user defined element)"
	elsif stack[2][1] != '@' then
		err = 1
		errtxt = "invalid user defined element name (must start with @)"
	elsif length(ude) = 0 then
		err = 1
		errtxt = "there are no user defined elements"
	end if
	if err then return end if

	i = 0
	stack[2] = " " & stack[2] & " "
	for j = 1 to length(ude) do
		if equal(stack[2], ude[j][1]) then
			ude = ude[1..j-1] & ude[j+1..$]
			i = 1
			exit
		end if
	end for
	if not i then
		err = 1
		errtxt = "user defined element" & stack[2] & "not found"
	end if
end procedure


global function ccopy(object f, object c)
-- copies content of file to the clipboard
-- c is copy command
-- returns 1 if file found, 0 if not found
	if length(c) = 0 then
		disp("Copy to clipboard command missing")
		return 0
	else
		if get_file_type(f) = 1 then
			system(c & " " & f, 2)
			return 1
		else
			return 0
		end if
	end if
end function


global function notmute()
-- returns true if not loop, or if loop with ? option (not flag_repeatmute)
-- some command messages not shown in loops, unless "do/repeat n ?"
	integer s
	s = 0
	if num_repeat = 0 then
		s = 1
	elsif flag_repeatmute = 0 then
		s = 1
	end if
	return s
end function


global function ndrand(atom mean, atom sd)
	-- Marsaglia polar method
	atom x1, x2, w, r
	integer flag_w
	flag_w = 0

	while flag_w = 0 do
		x1 = rnd() * 2 - 1
		x2 = rnd() * 2 - 1
		w = x1 * x1 + x2 * x2
		if w > 0 and w < 1 then flag_w = 1 end if
		
		if flag_w then
			if rnd() > 0.5 then x1 = x2 end if           -- randomly chose either first or second random number
			r = x1 * sqrt(log(w) * -2 / w)
			if r < -6 or r > 6 then flag_w = 0 end if    -- limit results to within -6 to +6 sigma
		end if
	end while

	r = r * sd + mean
	return r
end function


global procedure clearhy(integer mode)
	-- 0 = add line break
	-- 1 = clear
	integer D           -- file number

	if mode then
		D = open(hdir & "hy", "w")
		if D = -1 then
			write(2, "Warning: calculation result file hy could not be opened")
		else
			close(D)
		end if
	else
		D = open(hdir & "hy", "a")
		if D = -1 then
			write(2, "Warning: calculation result file hy could not be opened")
		else
			puts(D, "\n")
			close(D)
		end if
	end if
end procedure

global procedure savehy()
	-- saves result to file hy after each calculation

	object hyd          -- dir(hy)
	string rft          -- formatted result
	integer flag_w      -- open hy in "w" mode despite accumulation mode, because hy is empty
	integer D           -- file number
	integer lb          -- number of lines in buffer
	integer e           -- temporary flag variable

	if flag_silent then return end if

	rft = fmtresult()

	if flag_buffer = 0 then
		if flag_acc then
			flag_w = 0
			hyd = dir(hdir & "hy")
			if hyd = -1 then
				D = open(hdir & "hy", "w")
				flag_w = 1
			else
				if hyd[1][3] = 0 then                  -- open hy in "w" mode despite accumulation mode, because hy is empty
					D = open(hdir & "hy", "w")
					flag_w = 1
				else
					D = open(hdir & "hy", "a")
				end if
			end if

			if D = -1 then
				write(2, "Warning: calculation result file hy could not be opened")
			else
				if flag_w then
					puts(D, rft)                -- hy is empty, write result without preceding space or new line
				else
					if flag_acc = 1 then
						if flag_newlinehy then
							puts(D, rft)
						else
							puts(D, " " & rft)
						end if
					else
						puts(D, "\n" & rft)
					end if
				end if
				close(D)
			end if
		else
			D = open(hdir & "hy", "w")
			if D = -1 then
				disp("Warning: calculation result file hy could not be opened for writing")
			else
				puts(D, rft)
				close(D)
			end if
		end if
		
		if mode_copy then                         -- copyall is on
			if flag_run = 0 then                       -- not a script
				if num_repeat < 2 then                      -- not a loop (but can be repeat)
					e = ccopy(hdir & "hy", ccc)
					if e = 0 then
						write(2, "Warning: Copyall failed, couldn't find calculation result file hy")
					end if
				end if
			end if
		end if

	else                                        -- store result in buffer
		if flag_acc then
			if buffer = "" then
				buffer = append(buffer, rft)
			else
				if flag_acc = 2 then
					buffer = append(buffer, rft)
				else
					lb = length(buffer)
					if flag_newlinehy then
						buffer[lb] = buffer[lb] & rft
					else
						buffer[lb] = buffer[lb] & " " & rft
					end if
				end if
			end if
		else
			buffer = ""
			buffer = append(buffer, rft)
		end if
	end if
	flag_newlinehy = 0
end procedure


global procedure savehyin(object li)
	-- saves input line to file hyin after each input

	if li[$] = '?' then             -- ? at end of input line is removed
		li = atrim(li[1..$-1])
	end if
	integer D -- file number
	D = open(hdir & "hyin", "w")
	if D = -1 then
		disp("Warning: recent input line file hyin could not be opened for writing")
		flag_do = -1                 -- pass error condition to the DO code
	else
		puts(D, li)
		close(D)
	end if
end procedure


global procedure createini()
	-- creates empty hy.ini file if it doesn't exist

	integer D -- file number
	D = open(hdir & "hy.ini", "w")
	if D = -1 then
		disp("Warning: file hy.ini could not be created")
		disp("Make sure that Hypatia can write to its program folder!")
	else
		puts(D, "")
		close(D)
		disp("hy.ini file created in program folder -- you can edit it according to your needs\n")
		dispversion()
		dispbasics()
		disp("The above text is shown at first start of Hypatia (when no hy.ini file is found)")
		disp("or when you press ENTER with an empty input line\n")
	end if
end procedure


global procedure expand()
	-- inserts () files and user defined elements into input line op

	integer c -- counter
	integer l1, l2
	object f -- include file name
	object ff -- full file name
	object e, m
	object in -- include file data
	object opx -- receives repl() result
	integer cols -- number of screen columns
	integer flag_disp -- 1 = show line with inserted data
	integer k -- temp
	integer ops
	object opstrip -- part of op that gets stripped from op so that it won't be expanded
	c = 0 -- number of inserts found in line

-- lines beginning with help del del@ show show@ save save@ prompt buffer run edit exteditor cccommand do not get here
-- from lines that get here, everything after # has been removed -- so, comments do not get expanded
-- this still leaves in-command comments after SAVE, SAVE@, BUFFER SAVE and PROMPT in I1: I*: IF...ELSE ALSO: and ELSE: lines
-- therefore, everything after (and including) SAVE SAVE@ and PROMPT must not be expanded

	opstrip = ""
	ops = match(" save", op)
	if ops = 0 then ops = match(" prompt", op) end if
	if ops then
		opstrip = op[ops..$]
		op = op[1..ops-1]
	end if

	l1 = find('(', op)
	while l1 do
		c = c + 1
		if c > 10 then
			err = 1
			errtxt = "max. 10 inserts allowed in input line"
			exit
		end if

		l2 = find(')', op)
		if l2 < l1+2 then
			err = 1
			errtxt = ") missing or not at a valid position"
			exit
		else
			f = op[l1+1..l2-1]
			if equal(f, "buffer") then
				if flag_buffer then
					if length(buffer) then
						m = ""
						for j = 1 to length(buffer) do
							m = m & " " & buffer[j]
						end for
						m = m[2..$]
					else
						err = 1
						errtxt = "buffer is empty"
						exit
					end if
				else
					err = 1
					errtxt = "(buffer) needs buffer mode to be ON"
					exit
				end if
			else
				ff = hdir & f
				if get_file_type(ff) = 1 then
					in = readfile(ff)
					in = atrim(in)
					if length(in) = 0 then
						err = 1
						errtxt = "file " & f & " is empty"
						exit
					end if
					m = ""
					for j = 1 to length(in) do
						if length(in[j]) > 0 then
							if not begins("#", in[j]) then
								if length(m) then
									m = m & " " & in[j]
								else
								 	m = in[j]
								end if
							end if
						end if
					end for
				else
					err = 1
					errtxt = "file " & f & " not found"
					exit
				end if
			end if
		end if

		if l1 > 1 then e = op[1..l1-1] & " " else e = "" end if
		e = e & m
		if l2 < length(op) then e = e & " " & op[l2+1..$] end if
		op = e
		l1 = find('(', op)
	end while

	if c then
		k = match("  ", op)	                        -- remove double spaces after including insert files
		while k do
			op = op[1..k] & op[k+2..$]
			k = match("  ", op)
		end while
	end if

	if match('@', op) then                        -- expand user defined elements
		if match(" = ", op) then                       -- in an assign statement line the first word will not be expanded
			op = op & " "                                -- stored UDE names and values are enclosed in blanks, therefore match only when UDE name in input line is enclosed in blanks, too
			op = repl(op, "then ", "then")[2]           -- this also neeeds to apply after IF ... THEN
		else                                           --  will be replaced with space after the expansion loop.  is not an ASCII character and can therefore not be part of an input line
			op = " " & op & " "                          -- something like IF @ude N 10 >> THEN @ude = ... is allowed, UFE will be expanded in IF condition, but not in assign command
		end if
		for u = 1 to length(ude) do
			opx = repl(op, ude[u][1], ude[u][2])
			op = opx[2]
			c = c + opx[1]
		end for
		op = repl(op, "", " ")[2]
		op = atrim(op)
	end if

	if ops then op = op & opstrip end if         -- opx is returned to op after it was removed following save save@ or prompt

	if c then
		if num_repeat = 0 then                     -- always show expanded line when not a loop
			flag_disp = 1
		else                                       -- when a loop:
			flag_disp = 0
			if flag_repeatmute = 0 then              -- show expanded line when ? in loop command
				flag_disp = 1
			elsif flag_do and i_repeat = 1 then      -- first pass in DO loop
				if flag_runmute = 0 then               -- show expanded line with RUN file, not with _file
					flag_disp = 1
				end if
			end if
		end if
		
		if err = 0 then
			if flag_disp then
				-- limit display to screen width
	
				if flag_x then
					cols = 80                   -- *** video_config() currently doesn't work with Linux, 80 columns assumed ***
				else
					cols = video_config()[10]
				end if
	
				if length(op) < cols - 2 then
					disp("  " & op)
				else
					disp("  " & op[1..cols-6] & "...")
				end if
				if mode_log and err = 0 then write(0, "? " & op) end if
			end if
		end if
	end if

end procedure